/*
    vec_math.c - basic vector types and operations
    
    Copyright (C) 2009  Laurent Poirrier
    This file is part of the lin-kd-tree library.

    The lin-kd-tree library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, version 3 of the License.

    The lin-kd-tree library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <math.h>
#include <string.h>
#include <stdio.h>
#include "vec_math.h"

#define DEBUG_VAR(s...)	fprintf(stderr, s)

#define SCALAR_FORMAT	"%16g"

// *******************************************************************
// vec3 op
// *******************************************************************
void   vec3_set(vec3 *dst, const vec3 *src)
{
	dst->d[0] = src->d[0];
	dst->d[1] = src->d[1];
	dst->d[2] = src->d[2];
}

void   vec3_set_all(vec3 *dst, const scalar x)
{
	dst->d[0] = x;
	dst->d[1] = x;
	dst->d[2] = x;
}

void   vec3_set_scalars(vec3 *dst, const scalar x, const scalar y,
	const scalar z)
{
	dst->d[0] = x;
	dst->d[1] = y;
	dst->d[2] = z;
}

void   vec3_set_scaled(vec3 *dst, const vec3 *src, const scalar x)
{
	dst->d[0] = src->d[0] * x;
	dst->d[1] = src->d[1] * x;
	dst->d[2] = src->d[2] * x;
}

void   vec3_add(vec3 *dst, const vec3 *src)
{
	dst->d[0] += src->d[0];
	dst->d[1] += src->d[1];
	dst->d[2] += src->d[2];
}

void   vec3_add_scaled(vec3 *dst, const vec3 *src, const scalar x)
{
	dst->d[0] += src->d[0] * x;
	dst->d[1] += src->d[1] * x;
	dst->d[2] += src->d[2] * x;
}

void   vec3_sub(vec3 *dst, const vec3 *src)
{
	dst->d[0] -= src->d[0];
	dst->d[1] -= src->d[1];
	dst->d[2] -= src->d[2];
}

void   vec3_neg(vec3 *dst)
{
	dst->d[0] = -dst->d[0];
	dst->d[1] = -dst->d[1];
	dst->d[2] = -dst->d[2];
}

void   vec3_sum(vec3 *dst, const vec3 *a, const vec3 *b)
{
	dst->d[0] = a->d[0] + b->d[0];
	dst->d[1] = a->d[1] + b->d[1];
	dst->d[2] = a->d[2] + b->d[2];
}

void   vec3_scaled_sum(vec3 *dst, const vec3 *a,
	const vec3 *b, const scalar bx)
{
	dst->d[0] = a->d[0] + b->d[0] * bx;
	dst->d[1] = a->d[1] + b->d[1] * bx;
	dst->d[2] = a->d[2] + b->d[2] * bx;
}

void   vec3_weighted_sum(vec3 *dst, const vec3 *a, const scalar ax,
	const vec3 *b, const scalar bx)
{
	dst->d[0] = a->d[0] * ax + b->d[0] * bx;
	dst->d[1] = a->d[1] * ax + b->d[1] * bx;
	dst->d[2] = a->d[2] * ax + b->d[2] * bx;
}

void   vec3_difference(vec3 *dst, const vec3 *a, const vec3 *b)
{
	dst->d[0] = a->d[0] - b->d[0];
	dst->d[1] = a->d[1] - b->d[1];
	dst->d[2] = a->d[2] - b->d[2];
}

scalar vec3_sq_distance(const vec3 *a, const vec3 *b)
{
	vec3 v;

	v.d[0] = a->d[0] - b->d[0];
	v.d[1] = a->d[1] - b->d[1];
	v.d[2] = a->d[2] - b->d[2];
	return v.d[0] * v.d[0] + v.d[1] * v.d[1] + v.d[2] * v.d[2];
}

scalar vec3_distance(const vec3 *a, const vec3 *b)
{
	vec3 v;

	v.d[0] = a->d[0] - b->d[0];
	v.d[1] = a->d[1] - b->d[1];
	v.d[2] = a->d[2] - b->d[2];
	return sqrt(v.d[0] * v.d[0] + v.d[1] * v.d[1] + v.d[2] * v.d[2]);
}

void   vec3_scale(vec3 *dst, const scalar x)
{
	dst->d[0] *= x;
	dst->d[1] *= x;
	dst->d[2] *= x;
}

scalar vec3_sq_norm(const vec3 *v)
{
	return v->d[0] * v->d[0] + v->d[1] * v->d[1] + v->d[2] * v->d[2];
}

scalar vec3_norm(const vec3 *v)
{
	return sqrt(v->d[0] * v->d[0] + v->d[1] * v->d[1] + v->d[2] * v->d[2]);
}

void   vec3_normalize(vec3 *dst)
{
	// to develop, or gcc inlines it (or not critical) ?
	vec3_scale(dst, 1.0/vec3_norm(dst));
}

scalar vec3_norm_1(const vec3 *v)
{
	return v->d[0] + v->d[1] + v->d[2];
}

void   vec3_center(vec3 *dst, const vec3 *a, const vec3 *b)
{
	dst->d[0] = (a->d[0] + b->d[0]) * 0.5;
	dst->d[1] = (a->d[1] + b->d[1]) * 0.5;
	dst->d[2] = (a->d[2] + b->d[2]) * 0.5;
}

void   vec3_rotate_z(vec3 *dst, const vec3 *src, const angle a)
{
	scalar cosa, sina;
	
	sina = sin(a);
	cosa = cos(a);
	
	dst->d[0] = cosa * src->d[0] - sina * src->d[1];
	dst->d[1] = sina * src->d[0] + cosa * src->d[1];
}

void   vec3_rotate_x(vec3 *dst, const vec3 *src, const angle a)
{
	scalar cosa, sina;
	
	sina = sin(a);
	cosa = cos(a);
	
	dst->d[1] = cosa * src->d[1] - sina * src->d[2];
	dst->d[2] = sina * src->d[1] + cosa * src->d[2];
}

void   vec3_rotate_y(vec3 *dst, const vec3 *src, const angle a)
{
	scalar cosa, sina;
	
	sina = sin(a);
	cosa = cos(a);
	
	dst->d[0] =  cosa * src->d[0] + sina * src->d[2];
	dst->d[2] = -sina * src->d[0] + cosa * src->d[2];
}

// *******************************************************************
// matrix op
// *******************************************************************
void   mat3_set(mat3 *dst, const mat3 *a)
{
	memcpy(dst->d, a->d, 9*sizeof(scalar));
}

void   mat3x4_set(mat3x4 *dst, const mat3x4 *a)
{
	memcpy(dst->d, a->d, 12*sizeof(scalar));
}

void   mat4_set(mat4 *dst, const mat4 *a)
{
	memcpy(dst->d, a->d, 16*sizeof(scalar));
}

void   mat3_set_all(mat3 *dst, const scalar x)
{
	int t;
	for (t = 0; t < 9; t++) dst->d[t] = x;
}

void   mat3x4_set_all(mat3x4 *dst, const scalar x)
{
	int t;
	for (t = 0; t < 12; t++) dst->d[t] = x;
}

void   mat4_set_all(mat4 *dst, const scalar x)
{
	int t;
	for (t = 0; t < 16; t++) dst->d[t] = x;
}


void   mat3_identity(mat3 *dst)
{
	dst->d[1] = dst->d[2] = dst->d[3] = dst->d[5] = dst->d[6] =
		dst->d[7] = 0.0;
	dst->d[0] = dst->d[4] = dst->d[8] = 1.0;
}

void   mat3x4_identity(mat3x4 *dst)
{
	dst->d[1] = dst->d[2] = dst->d[3] = dst->d[4] = dst->d[6] =
		dst->d[7] = dst->d[8] = dst->d[9] = dst->d[11] = 0.0;
	dst->d[0] = dst->d[5] = dst->d[10] = 1.0;
}

void   mat4_identity(mat4 *dst)
{
	dst->d[1] = dst->d[2] = dst->d[3] = dst->d[4] = dst->d[6] =
		dst->d[7] = dst->d[8] = dst->d[9] = dst->d[11] =
		dst->d[12] = dst->d[13] = dst->d[14] = 0.0;
	dst->d[0] = dst->d[5] = dst->d[10] = dst->d[15] = 1.0;
}


void   mat3_scale(mat3 *dst, const scalar x)
{
	int t;
	for (t = 0; t < 9; t++) dst->d[t] *= x;
}

void   mat3x4_scale(mat3x4 *dst, const scalar x)
{
	int t;
	for (t = 0; t < 9; t++) dst->d[t] *= x;
}

void   mat4_scale(mat4 *dst, const scalar x)
{
	int t;
	for (t = 0; t < 9; t++) dst->d[t] *= x;
}

// *******************************************************************
// sum of matrices, transposes
// *******************************************************************
void   mat3_add(mat3 *dst, const mat3 *a)
{
	int t;
	
	for (t = 0; t < 9; t++) dst->d[t] += a->d[t];
}

void   mat3_neg(mat3 *dst)
{
	int t;
	
	for (t = 0; t < 9; t++) dst->d[t] = -dst->d[t];
}

void   mat3_add_scaled_identity(mat3 *dst, const scalar x)
{
	dst->d[0] += x;
	dst->d[4] += x;
	dst->d[8] += x;
}

#define MAT_SWAP(m, i, j)	tmp = m[i]; m[i] = m[j]; m[j] = tmp

void   mat3_inplace_transpose(mat3 *dst)
{
	scalar tmp;
	
	MAT_SWAP(dst->d, 1, 3);
	MAT_SWAP(dst->d, 2, 6);
	MAT_SWAP(dst->d, 5, 7);
}

void   mat3x4_3x3_inplace_transpose(mat4 *dst)
{
	scalar tmp;
	
	MAT_SWAP(dst->d, 1, 4);
	MAT_SWAP(dst->d, 2, 8);
	MAT_SWAP(dst->d, 6, 9);
}

void   mat4_inplace_transpose(mat4 *dst)
{
	scalar tmp;
	
	MAT_SWAP(dst->d, 1, 4);
	MAT_SWAP(dst->d, 2, 8);
	MAT_SWAP(dst->d, 3, 12);
	MAT_SWAP(dst->d, 6, 9);
	MAT_SWAP(dst->d, 7, 13);
	MAT_SWAP(dst->d, 11, 14);
}


void   mat3_transpose(mat3 *dst, const mat3 *a)
{
	dst->d[0] = a->d[0];	dst->d[1] = a->d[3];	dst->d[2] = a->d[6];
	dst->d[3] = a->d[1];	dst->d[4] = a->d[4];	dst->d[5] = a->d[7];
	dst->d[6] = a->d[2];	dst->d[7] = a->d[5];	dst->d[8] = a->d[8];
}

void   mat3x4_3x3_transpose(mat3x4 *dst, const mat3x4 *a)
{
	dst->d[0] = a->d[0];	dst->d[1] = a->d[4];	dst->d[2] = a->d[8];
	dst->d[4] = a->d[1];	dst->d[5] = a->d[5];	dst->d[6] = a->d[9];
	dst->d[8] = a->d[2];	dst->d[9] = a->d[6];	dst->d[10] = a->d[10];
}

void   mat3x4_transpose(mat3x4 *dst, const mat3x4 *a)
{
	dst->d[0] = a->d[0];
	dst->d[1] = a->d[4];
	dst->d[2] = a->d[8];
	dst->d[3] = 0.0;
	dst->d[4] = a->d[1];
	dst->d[5] = a->d[5];
	dst->d[6] = a->d[9];
	dst->d[7] = 0.0;
	dst->d[8] = a->d[2];
	dst->d[9] = a->d[6];
	dst->d[10] = a->d[10];
	dst->d[11] = 0.0;
}

void   mat4_transpose(mat4 *dst, const mat4 *a)
{
	dst->d[0] = a->d[0];
	dst->d[1] = a->d[4];
	dst->d[2] = a->d[8];
	dst->d[3] = a->d[12];
	dst->d[4] = a->d[1];
	dst->d[5] = a->d[5];
	dst->d[6] = a->d[9];
	dst->d[7] = a->d[13];
	dst->d[8] = a->d[2];
	dst->d[9] = a->d[6];
	dst->d[10] = a->d[10];
	dst->d[11] = a->d[14];
	dst->d[12] = a->d[3];
	dst->d[13] = a->d[7];
	dst->d[14] = a->d[11];
	dst->d[15] = a->d[15];
}

void   mat4_mat3x4_transpose(mat4 *dst, const mat3x4 *a)
{
	dst->d[0] = a->d[0];
	dst->d[1] = a->d[4];
	dst->d[2] = a->d[8];
	dst->d[3] = 0.0;
	dst->d[4] = a->d[1];
	dst->d[5] = a->d[5];
	dst->d[6] = a->d[9];
	dst->d[7] = 0.0;
	dst->d[8] = a->d[2];
	dst->d[9] = a->d[6];
	dst->d[10] = a->d[10];
	dst->d[11] = 0.0;
	dst->d[12] = a->d[3];
	dst->d[13] = a->d[7];
	dst->d[14] = a->d[11];
	dst->d[15] = 1.0;
}

// *******************************************************************
// vec3 skew-symmetric matrix
// *******************************************************************
void   vec3_skew_sym(mat3 *dst, const vec3 *src)
{
	dst->d[0] = dst->d[4] = dst->d[8] = 0.0;
	
	dst->d[1] = -src->d[2];
	dst->d[3] = src->d[2];
	
	dst->d[2] = src->d[1];
	dst->d[6] = -src->d[1];
	
	dst->d[5] = -src->d[0];
	dst->d[7] = src->d[0];
}

void   vec3_sq_skew_sym(mat3 *dst, const vec3 *src)
{
	vec3 td;
	
	td.d[0] = -src->d[0] * src->d[0];
	td.d[1] = -src->d[1] * src->d[1];
	td.d[2] = -src->d[2] * src->d[2];
	
	dst->d[0] = td.d[1] + td.d[2];
	dst->d[4] = td.d[0] + td.d[2];
	dst->d[8] = td.d[0] + td.d[1];
	
	dst->d[1] = dst->d[3] = src->d[0] * src->d[1];
	dst->d[2] = dst->d[6] = src->d[0] * src->d[2];
	dst->d[5] = dst->d[7] = src->d[1] * src->d[2];
}


// *******************************************************************
// transformation matrices
// *******************************************************************
void   mat3x4_translation(mat3x4 *dst, const vec3 *d)
{
	dst->d[1] = dst->d[2] = dst->d[4] =
		dst->d[6] = dst->d[8] = dst->d[9] = 0.0;
	dst->d[0] = dst->d[5] = dst->d[10] = 1.0;
	
	dst->d[3] = d->d[0];
	dst->d[7] = d->d[1];
	dst->d[11] = d->d[2];
}

void   mat3x4_translation_scalars(mat3x4 *dst,
	const scalar x, const scalar y, const scalar z)
{
	dst->d[1] = dst->d[2] = dst->d[4] =
		dst->d[6] = dst->d[8] = dst->d[9] = 0.0;
	dst->d[0] = dst->d[5] = dst->d[10] = 1.0;
	
	dst->d[3] = x;
	dst->d[7] = y;
	dst->d[11] = z;
}

void   mat3x4_set_translation(mat3x4 *dst, const vec3 *d)
{
	dst->d[3] = d->d[0];
	dst->d[7] = d->d[1];
	dst->d[11] = d->d[2];
}

void   mat3x4_rotation(mat3x4 *dst, const vec3 *d, const angle a)
{
	scalar _1mc, s;
	scalar sx, sy, sz;
	scalar x2, y2, z2;
	scalar mxy, myz, mxz;
	
	s = sin(a);
	_1mc = 1.0 - cos(a);
	
	sx = s * dst->d[0];
	sy = s * dst->d[1];
	sz = s * dst->d[2];
	
	x2 = dst->d[0] * dst->d[0];
	y2 = dst->d[1] * dst->d[1];
	z2 = dst->d[2] * dst->d[2];
	
	mxy = - _1mc * dst->d[0] * dst->d[1];
	myz = - _1mc * dst->d[1] * dst->d[2];
	mxz = - _1mc * dst->d[2] * dst->d[0];
	
	dst->d[0] = 1.0 + _1mc * (y2 + z2);
	dst->d[1] = mxy - sz;
	dst->d[2] = mxz + sy;
	dst->d[3] = 0.0;
	dst->d[4] = mxy + sz;
	dst->d[5] = 1.0 + _1mc * (x2 + z2);
	dst->d[6] = myz - sx;
	dst->d[7] = 0.0;
	dst->d[8] = mxz - sy;
	dst->d[9] = myz + sx;
	dst->d[10] = 1.0 + _1mc * (x2 + y2);
	dst->d[11] = 0.0;
}

void   mat3x4_rotation_scalars(mat3x4 *dst,
	const scalar x, const scalar y, const scalar z, const angle a)
{
	vec3 td;
	
	vec3_set_scalars(&td, x, y, z);
	mat3x4_rotation(dst, &td, a);
}

void   mat3x4_rotation_z(mat3x4 *dst, const angle a)
{
	scalar cosa, sina;
	
	sina = sin(a);
	cosa = cos(a);
	
	dst->d[0] = cosa; dst->d[1] = -sina; dst->d[2] = 0.0; dst->d[3] = 0.0;
	dst->d[4] = sina; dst->d[5] = cosa;  dst->d[6] = 0.0; dst->d[7] = 0.0;
	dst->d[8] = 0.0;  dst->d[9] = 0.0;  dst->d[10] = 1.0; dst->d[11] = 0.0;
}


void   mat3x4_rotation_x(mat3x4 *dst, const angle a)
{
	scalar cosa, sina;
	
	sina = sin(a);
	cosa = cos(a);
	
	dst->d[0] = 1.0; dst->d[1] = 0.0;  dst->d[2] = 0.0;   dst->d[3] = 0.0;
	dst->d[4] = 0.0; dst->d[5] = cosa; dst->d[6] = -sina; dst->d[7] = 0.0;
	dst->d[8] = 0.0; dst->d[9] = sina; dst->d[10] = cosa; dst->d[11] = 0.0;
}


void   mat3x4_rotation_y(mat3x4 *dst, const angle a)
{
	scalar cosa, sina;
	
	sina = sin(a);
	cosa = cos(a);
	
	dst->d[0] = cosa;  dst->d[1] = 0.0; dst->d[2] = sina;  dst->d[3] = 0.0;
	dst->d[4] = 0.0;   dst->d[5] = 1.0; dst->d[6] = 0.0;   dst->d[7] = 0.0;
	dst->d[8] = -sina; dst->d[9] = 0.0; dst->d[10] = cosa; dst->d[11] = .0;
}


void   mat3x4_rotation_axis(mat3x4 *dst, const int axis, const angle a)
{
	switch(axis) {
	case 0:
		mat3x4_rotation_x(dst, a);
		return;
	case 1:
		mat3x4_rotation_y(dst, a);
		return;
	case 2:
		mat3x4_rotation_z(dst, a);
		return;
	default:
		break;
	}
}


// *******************************************************************
// vector products
// *******************************************************************
scalar vec3_dot_product(const vec3 *a, const vec3 *b)
{
	return a->d[0] * b->d[0] + a->d[1] * b->d[1] + a->d[2] * b->d[2];
}

void   vec3_cross_product(vec3 *dst, const vec3 *a, const vec3 *b)
{
	dst->d[0] = a->d[1] * b->d[2] - a->d[2] * b->d[1];
	dst->d[1] = a->d[2] * b->d[0] - a->d[0] * b->d[2];
	dst->d[2] = a->d[0] * b->d[1] - a->d[1] * b->d[0];
}

// *******************************************************************
// matrix-vector products
// *******************************************************************
void   mat3_vec3_product(vec3 *dst, const mat3 *a, const vec3 *v)
{
	dst->d[0] = a->d[0] * v->d[0] + a->d[1] * v->d[1] + a->d[2] * v->d[2];
	dst->d[1] = a->d[3] * v->d[0] + a->d[4] * v->d[1] + a->d[5] * v->d[2];
	dst->d[2] = a->d[6] * v->d[0] + a->d[7] * v->d[1] + a->d[8] * v->d[2];
}

void   mat3x4_vec3_product(vec3 *dst, const mat3x4 *a, const vec3 *v)
{
	dst->d[0] = a->d[0] * v->d[0] + a->d[1] * v->d[1] + a->d[2] * v->d[2]
		+ a->d[3];
	dst->d[1] = a->d[4] * v->d[0] + a->d[5] * v->d[1] + a->d[6] * v->d[2]
		+ a->d[7];
	dst->d[2] = a->d[8] * v->d[0] + a->d[9] * v->d[1] + a->d[10] * v->d[2]
		+ a->d[11];
}

void   mat3x4_3x3_vec3_product(vec3 *dst, const mat3x4 *a, const vec3 *v)
{
	dst->d[0] = a->d[0] * v->d[0] + a->d[1] * v->d[1] + a->d[2] * v->d[2];
	dst->d[1] = a->d[4] * v->d[0] + a->d[5] * v->d[1] + a->d[6] * v->d[2];
	dst->d[2] = a->d[8] * v->d[0] + a->d[9] * v->d[1] + a->d[10] * v->d[2];
}

void   mat3x4_column_0(vec3 *dst, const mat3x4 *a)
{
	dst->d[0] = a->d[0];
	dst->d[1] = a->d[4];
	dst->d[2] = a->d[8];
}

void   mat3x4_column_1(vec3 *dst, const mat3x4 *a)
{
	dst->d[0] = a->d[1];
	dst->d[1] = a->d[5];
	dst->d[2] = a->d[9];
}

void   mat3x4_column_2(vec3 *dst, const mat3x4 *a)
{
	dst->d[0] = a->d[2];
	dst->d[1] = a->d[6];
	dst->d[2] = a->d[10];
}

void   mat3x4_column_3(vec3 *dst, const mat3x4 *a)
{
	dst->d[0] = a->d[3];
	dst->d[1] = a->d[7];
	dst->d[2] = a->d[11];
}

void   mat3x4_3x3_line_0(vec3 *dst, const mat3x4 *a)
{
	dst->d[0] = a->d[0];
	dst->d[1] = a->d[1];
	dst->d[2] = a->d[2];
}

void   mat3x4_3x3_line_1(vec3 *dst, const mat3x4 *a)
{
	dst->d[0] = a->d[4];
	dst->d[1] = a->d[5];
	dst->d[2] = a->d[6];
}

void   mat3x4_3x3_line_2(vec3 *dst, const mat3x4 *a)
{
	dst->d[0] = a->d[8];
	dst->d[1] = a->d[9];
	dst->d[2] = a->d[10];
}


// *******************************************************************
// matrix products
// *******************************************************************
#define MAT3_LINE(d, s, a, b, l) \
	d[l  ] s a[l  ] * b[0] + a[l+1] * b[3] + a[l+2] * b[6]; \
	d[l+1] s a[l  ] * b[1] + a[l+1] * b[4] + a[l+2] * b[7]; \
	d[l+2] s a[l  ] * b[2] + a[l+1] * b[5] + a[l+2] * b[8]

void   mat3_product(mat3 *dst, const mat3 *a, const mat3 *b)
{
	MAT3_LINE(dst->d, =, a->d, b->d, 0);
	MAT3_LINE(dst->d, =, a->d, b->d, 3);
	MAT3_LINE(dst->d, =, a->d, b->d, 6);
}

void   mat3_add_product(mat3 *dst, const mat3 *a, const mat3 *b)
{
	MAT3_LINE(dst->d, +=, a->d, b->d, 0);
	MAT3_LINE(dst->d, +=, a->d, b->d, 3);
	MAT3_LINE(dst->d, +=, a->d, b->d, 6);
}

void   mat3_sub_product(mat3 *dst, const mat3 *a, const mat3 *b)
{
	MAT3_LINE(dst->d, -=, a->d, b->d, 0);
	MAT3_LINE(dst->d, -=, a->d, b->d, 3);
	MAT3_LINE(dst->d, -=, a->d, b->d, 6);
}

#define MAT3X4_LINE(d, a, b, l) \
	d[l  ] = a[l  ] * b[0] + a[l+1] * b[4] + a[l+2] * b[8]; \
	d[l+1] = a[l  ] * b[1] + a[l+1] * b[5] + a[l+2] * b[9]; \
	d[l+2] = a[l  ] * b[2] + a[l+1] * b[6] + a[l+2] * b[10]; \
	d[l+3] = a[l  ] * b[3] + a[l+1] * b[7] + a[l+2] * b[11] + a[l+3]

void   mat3x4_product(mat3x4 *dst, const mat3x4 *a, const mat3x4 *b)
{
	MAT3X4_LINE(dst->d, a->d, b->d, 0);
	MAT3X4_LINE(dst->d, a->d, b->d, 4);
	MAT3X4_LINE(dst->d, a->d, b->d, 8);
}

#define MAT4_LINE(d, a, b, l) \
 d[l  ] = a[l  ] * b[0] + a[l+1] * b[4] + a[l+2] * b[8] + a[l+3] * b[12]; \
 d[l+1] = a[l  ] * b[1] + a[l+1] * b[5] + a[l+2] * b[9] + a[l+3] * b[13]; \
 d[l+2] = a[l  ] * b[2] + a[l+1] * b[6] + a[l+2] * b[10] + a[l+3] * b[14]; \
 d[l+3] = a[l  ] * b[3] + a[l+1] * b[7] + a[l+2] * b[11] + a[l+3] * b[15]

void   mat4_product(mat4 *dst, const mat4 *a, const mat4 *b)
{
	MAT4_LINE(dst->d, a->d, b->d, 0);
	MAT4_LINE(dst->d, a->d, b->d, 4);
	MAT4_LINE(dst->d, a->d, b->d, 8);
	MAT4_LINE(dst->d, a->d, b->d, 12);
}



void   mat3_post_mult(mat3 *dst, const mat3 *b)
{
	mat3 tmp;
	
	mat3_product(&tmp, dst, b);
	mat3_set(dst, &tmp);
}

void   mat3x4_post_mult(mat3x4 *dst, const mat3x4 *b)
{
	mat3x4 tmp;
	
	mat3x4_product(&tmp, dst, b);
	mat3x4_set(dst, &tmp);
}

void   mat4_post_mult(mat4 *dst, const mat4 *b)
{
	mat4 tmp;
	
	mat4_product(&tmp, dst, b);
	mat4_set(dst, &tmp);
}

void   mat3_pre_mult(mat3 *dst, const mat3 *a)
{
	mat3 tmp;
	
	mat3_product(&tmp, a, dst);
	mat3_set(dst, &tmp);
}

void   mat3x4_pre_mult(mat3x4 *dst, const mat3x4 *a)
{
	mat3x4 tmp;
	
	mat3x4_product(&tmp, a, dst);
	mat3x4_set(dst, &tmp);
}

void   mat4_pre_mult(mat4 *dst, const mat4 *a)
{
	mat4 tmp;
	
	mat4_product(&tmp, a, dst);
	mat4_set(dst, &tmp);
}


#define MAT3_3X4_3X3_LINE(d, a, b, la, lb) \
	d[lb  ] = a[la  ] * b[0] + a[la+1] * b[3] + a[la+2] * b[6]; \
	d[lb+1] = a[la  ] * b[1] + a[la+1] * b[4] + a[la+2] * b[7]; \
	d[lb+2] = a[la  ] * b[2] + a[la+1] * b[5] + a[la+2] * b[8]

#define MAT3_3X3_3X4_TR_LINE(d, a, b, la) \
	d[la  ] = a[la  ] * b[0] + a[la+1] * b[1] + a[la+2] * b[2]; \
	d[la+1] = a[la  ] * b[4] + a[la+1] * b[5] + a[la+2] * b[6]; \
	d[la+2] = a[la  ] * b[8] + a[la+1] * b[9] + a[la+2] * b[10]

void   mat3_mat3x4_3x3_transform(mat3 *dst, const mat3x4 *m, const mat3 *a)
{
	mat3 ma;
	
	MAT3_3X4_3X3_LINE(ma.d, m->d, a->d, 0, 0);
	MAT3_3X4_3X3_LINE(ma.d, m->d, a->d, 4, 3);
	MAT3_3X4_3X3_LINE(ma.d, m->d, a->d, 8, 6);
	
	MAT3_3X3_3X4_TR_LINE(dst->d, ma.d, m->d, 0);
	MAT3_3X3_3X4_TR_LINE(dst->d, ma.d, m->d, 3);
	MAT3_3X3_3X4_TR_LINE(dst->d, ma.d, m->d, 6);
}


// *******************************************************************
// 2x2 & 3x3 determinant macro
// *******************************************************************
#define DET_2(a, a00, a01, a10, a11) \
	( a[a00] * a[a11] - a[a10] * a[a01] )

// 12 multiplications
#define DET_3(a, a00, a01, a02, a10, a11, a12, a20, a21, a22) \
	( a[a00] * a[a11] * a[a22] \
	+ a[a01] * a[a12] * a[a20] \
	+ a[a02] * a[a10] * a[a21] \
	- a[a02] * a[a11] * a[a20] \
	- a[a00] * a[a12] * a[a21] \
	- a[a01] * a[a10] * a[a22] )

// *******************************************************************
// 3x3 determinant & inverse
// *******************************************************************
// For small matrices, the Laplace formula using cofactors is faster.
// For gauss-jordan elimination algorithm, see matrix_math.

scalar mat3_determinant(const mat3 *src)
{
	return(DET_3(src->d,0,1,2,3,4,5,6,7,8));
}

void   mat3_determined_inverse(mat3 *dst, const mat3 *src, const scalar detsrc)
{
	scalar id;

	id = 1.0/detsrc;

	dst->d[0]  =  id*DET_2(src->d,4,5,7,8);
	dst->d[1]  = -id*DET_2(src->d,1,2,7,8);
	dst->d[2]  =  id*DET_2(src->d,1,2,4,5);
	dst->d[3]  = -id*DET_2(src->d,3,5,6,8);
	dst->d[4]  =  id*DET_2(src->d,0,2,6,8);
	dst->d[5]  = -id*DET_2(src->d,0,2,3,5);
	dst->d[6]  =  id*DET_2(src->d,3,4,6,7);
	dst->d[7]  = -id*DET_2(src->d,0,1,6,7);
	dst->d[8]  =  id*DET_2(src->d,0,1,3,4);
}

void  mat3_inverse(mat3 *dst, const mat3 *src)
{
	mat3_determined_inverse(dst, src, DET_3(src->d,0,1,2,3,4,5,6,7,8));
}


// *******************************************************************
// 4x3 determinant & inverse
// *******************************************************************
scalar mat3x4_determinant(const mat3x4 *src)
{
	return(DET_3(src->d,0,1,2,4,5,6,8,9,10));
}

void mat3x4_inverse(mat3x4 *dst, const mat3x4 *src)
{
	scalar id;

	id = 1.0/(DET_3(src->d,0,1,2,4,5,6,8,9,10));

	dst->d[0]  =  id*DET_2(src->d,5,6,9,10);
	dst->d[1]  = -id*DET_2(src->d,1,2,9,10);
	dst->d[2]  =  id*DET_2(src->d,1,2,5,6);
	dst->d[3]  = -id*DET_3(src->d,1,2,3,5,6,7,9,10,11);
	dst->d[4]  = -id*DET_2(src->d,4,6,8,10);
	dst->d[5]  =  id*DET_2(src->d,0,2,8,10);
	dst->d[6]  = -id*DET_2(src->d,0,2,4,6);
	dst->d[7]  =  id*DET_3(src->d,0,2,3,4,6,7,8,10,11);
	dst->d[8]  =  id*DET_2(src->d,4,5,8,9);
	dst->d[9]  = -id*DET_2(src->d,0,1,8,9);
	dst->d[10] =  id*DET_2(src->d,0,1,4,5);
	dst->d[11] = -id*DET_3(src->d,0,1,3,4,5,7,8,9,11);
	//dst->d[12] = dst->d[13] =  dst->d[14] = .0;
	//dst->d[15] = 1.0;
}

// *******************************************************************
// 3x4 [ R | t ] inverse
// *******************************************************************
void   mat3x4_Rt_inverse(mat3x4 *dst, const mat3x4 *src)
{
	// if R ^ (-1) == R', where R' is the transpose of R
	// then [ R | t ] ^ (-1) = [ R' | -R't ]
	
	dst->d[0] = src->d[0];
	dst->d[1] = src->d[4];
	dst->d[2] = src->d[8];
	dst->d[3] = - src->d[0] * src->d[3] - src->d[4] * src->d[7]
		- src->d[8] * src->d[11];
	dst->d[4] = src->d[1];
	dst->d[5] = src->d[5];
	dst->d[6] = src->d[9];
	dst->d[7] = - src->d[1] * src->d[3] - src->d[5] * src->d[7]
		- src->d[9] * src->d[11];
	dst->d[8] = src->d[2];
	dst->d[9] = src->d[6];
	dst->d[10] = src->d[10];
	dst->d[11] = - src->d[2] * src->d[3] - src->d[6] * src->d[7]
		- src->d[10] * src->d[11];
}


// *******************************************************************
// 4x4 determinant & inverse
// *******************************************************************

// 52 multiplications
scalar mat4_determinant(const mat4 *src)
{
	return( -src->d[12]*DET_3(src->d,1,2,3,5,6,7,9,10,11)
		+src->d[13]*DET_3(src->d,0,2,3,4,6,7,8,10,11)
		-src->d[14]*DET_3(src->d,0,1,3,4,5,7,8,9,11)
		+src->d[15]*DET_3(src->d,0,1,2,4,5,6,8,9,10) );
}

void mat4_inverse(mat4 *dst, const mat4 *src)
{
	scalar id;
	int t;

	dst->d[0]  =  DET_3(src->d,5,6,7,9,10,11,13,14,15);
	dst->d[1]  = -DET_3(src->d,1,2,3,9,10,11,13,14,15);
	dst->d[2]  =  DET_3(src->d,1,2,3,5,6,7,13,14,15);
	dst->d[3]  = -DET_3(src->d,1,2,3,5,6,7,9,10,11);
	dst->d[4]  = -DET_3(src->d,4,6,7,8,10,11,12,14,15);
	dst->d[5]  =  DET_3(src->d,0,2,3,8,10,11,12,14,15);
	dst->d[6]  = -DET_3(src->d,0,2,3,4,6,7,12,14,15);
	dst->d[7]  =  DET_3(src->d,0,2,3,4,6,7,8,10,11);
	dst->d[8]  =  DET_3(src->d,4,5,7,8,9,11,12,13,15);
	dst->d[9]  = -DET_3(src->d,0,1,3,8,9,11,12,13,15);
	dst->d[10] =  DET_3(src->d,0,1,3,4,5,7,12,13,15);
	dst->d[11] = -DET_3(src->d,0,1,3,4,5,7,8,9,11);
	dst->d[12] = -DET_3(src->d,4,5,6,8,9,10,12,13,14);
	dst->d[13] =  DET_3(src->d,0,1,2,8,9,10,12,13,14);
	dst->d[14] = -DET_3(src->d,0,1,2,4,5,6,12,13,14);
	dst->d[15] =  DET_3(src->d,0,1,2,4,5,6,8,9,10);

	id = 1.0/(
		src->d[0] * dst->d[0] + src->d[1] * dst->d[4]
		+ src->d[2] * dst->d[8] + src->d[3] * dst->d[12]);
	for (t=0;t<16;t++) dst->d[t] *= id;
}


// *******************************************************************
// Debug helpers
// *******************************************************************
#define FORMAT	SCALAR_FORMAT
#define FORMAT3	FORMAT FORMAT FORMAT
#define FORMAT4	FORMAT FORMAT FORMAT FORMAT
void scalar_print(const char *name, const scalar x)
{
	DEBUG_VAR("%s = " FORMAT "\n", name, x);
}

void vec3_print(const char *name, const vec3 *v)
{
	DEBUG_VAR("%s = \n" FORMAT3 "\n", name, v->d[0], v->d[1], v->d[2]);
}

void vec4_print(const char *name, const vec4 *v)
{
	DEBUG_VAR("%s = \n" FORMAT4 "\n",
		name, v->d[0], v->d[1], v->d[2], v->d[3]);
}

void mat3_print(const char *name, const mat3 *a)
{
	int i;
	
	DEBUG_VAR("%s = \n", name);
	for (i=0;i<9;i+=3)
		DEBUG_VAR(" " FORMAT3 "\n", a->d[i], a->d[i+1], a->d[i+2]);
}

void mat3x4_print(const char *name, const mat3x4 *a)
{
	int i;
	
	DEBUG_VAR("%s = \n", name);
	for (i=0;i<12;i+=4)
		DEBUG_VAR(" " FORMAT4 "\n",
			a->d[i], a->d[i+1], a->d[i+2], a->d[i+3]);
	DEBUG_VAR("[" FORMAT4 " ]\n", .0, .0, .0, 1.0);
}

void mat4_print(const char *name, const mat4 *a)
{
	int i;
	
	DEBUG_VAR("%s = \n", name);
	for (i=0;i<16;i+=4)
		DEBUG_VAR(" " FORMAT4 "\n",
			a->d[i], a->d[i+1], a->d[i+2], a->d[i+3]);
}


