/*
    vec_math.h - basic vector types and operations
    
    Copyright (C) 2009  Laurent Poirrier
    This file is part of the lin-kd-tree library.

    The lin-kd-tree library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, version 3 of the License.

    The lin-kd-tree library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __VECTOR_MATH_H
#define __VECTOR_MATH_H

/* Note:

Here, mat3x4 stands for 3-lines-by-4-columns matrices. This is
equivalent to OpenGL's mat4x3 type...

*/

// *******************************************************************
// Types
// *******************************************************************

#define VEC_GL(function)	function ## d
#define VEC_GL_V(function)	function ## dv

typedef double scalar;

typedef double angle;

typedef struct { scalar d[3];  } vec3;
typedef struct { scalar d[4];  } vec4;

typedef struct { scalar d[9];  } mat3x3;
typedef struct { scalar d[12]; } mat3x4;
typedef struct { scalar d[16]; } mat4x4;

typedef mat3x3 mat3;
typedef mat4x4 mat4;

// *******************************************************************
// Functions
// *******************************************************************

// [added as needs arise]

// vector basics
void   vec3_set(vec3 *dst, const vec3 *src);
void   vec3_set_scaled(vec3 *dst, const vec3 *src, const scalar x);
void   vec3_set_all(vec3 *dst, const scalar x);
void   vec3_set_scalars(vec3 *dst, const scalar x, const scalar y,
	const scalar z);
void   vec3_add(vec3 *dst, const vec3 *src);
void   vec3_add_scaled(vec3 *dst, const vec3 *src, const scalar x);
void   vec3_sub(vec3 *dst, const vec3 *src);
void   vec3_neg(vec3 *dst);

void   vec3_center(vec3 *dst, const vec3 *a, const vec3 *b);

void   vec3_sum(vec3 *dst, const vec3 *a, const vec3 *b);
void   vec3_scaled_sum(vec3 *dst, const vec3 *a,
	const vec3 *b, const scalar bx);
void   vec3_weighted_sum(vec3 *dst, const vec3 *a, const scalar ax,
	const vec3 *b, const scalar bx);
void   vec3_difference(vec3 *dst, const vec3 *a, const vec3 *b);
scalar vec3_sq_distance(const vec3 *a, const vec3 *b);
scalar vec3_distance(const vec3 *a, const vec3 *b);

void   vec3_scale(vec3 *dst, const scalar x);
scalar vec3_sq_norm(const vec3 *src);
scalar vec3_norm(const vec3 *src);
void   vec3_normalize(vec3 *dst);

scalar vec3_norm_1(const vec3 *src);

void   vec3_rotate_z(vec3 *dst, const vec3 *src, const angle a);
void   vec3_rotate_x(vec3 *dst, const vec3 *src, const angle a);
void   vec3_rotate_y(vec3 *dst, const vec3 *src, const angle a);

// matrix basics
void   mat3_set(mat3 *dst, const mat3 *a);
void   mat3x4_set(mat3x4 *dst, const mat3x4 *a);
void   mat4_set(mat4 *dst, const mat4 *a);

void   mat3_set_all(mat3 *dst, const scalar x);
void   mat3x4_set_all(mat3x4 *dst, const scalar x);
void   mat4_set_all(mat4 *dst, const scalar x);

void   mat3_identity(mat3 *dst);
void   mat3x4_identity(mat3x4 *dst);
void   mat4_identity(mat4 *dst);

void   mat3_scale(mat3 *dst, const scalar x);
void   mat3x4_scale(mat3x4 *dst, const scalar x);
void   mat4_scale(mat4 *dst, const scalar x);

void   mat3_add(mat3 *dst, const mat3 *a);
void   mat3_neg(mat3 *dst);
void   mat3_add_scaled_identity(mat3 *dst, const scalar x);

void   mat3_inplace_transpose(mat3 *dst);
void   mat3x4_3x3_inplace_transpose(mat4 *dst);
void   mat4_inplace_transpose(mat4 *dst);

void   mat3_transpose(mat3 *dst, const mat3 *a);
	// zeroes last column
void   mat3x4_transpose(mat3x4 *dst, const mat3x4 *a);
	// leaves last column unchanged (from what was in *dst)
void   mat3x4_3x3_transpose(mat3x4 *dst, const mat3x4 *a);
void   mat4_transpose(mat4 *dst, const mat4 *a);
void   mat4_mat3x4_transpose(mat4 *dst, const mat3x4 *a);

// vector's skew-symmetric matrices
void   vec3_skew_sym(mat3 *dst, const vec3 *src);
void   vec3_sq_skew_sym(mat3 *dst, const vec3 *src);

// transformation matrices
void   mat3x4_translation(mat3x4 *dst, const vec3 *d);
void   mat3x4_translation_scalars(mat3x4 *dst,
	const scalar x, const scalar y, const scalar z);
void   mat3x4_set_translation(mat3x4 *dst, const vec3 *d);
	// axis direction vectors must be normalized
void   mat3x4_rotation(mat3x4 *dst, const vec3 *d, const angle a);
void   mat3x4_rotation_scalars(mat3x4 *dst,
	const scalar x, const scalar y, const scalar z, const angle a);
void   mat3x4_rotation_z(mat3x4 *dst, const angle a);
void   mat3x4_rotation_x(mat3x4 *dst, const angle a);
void   mat3x4_rotation_y(mat3x4 *dst, const angle a);
void   mat3x4_rotation_axis(mat3x4 *dst, const int axis, const angle a);

// multiplications
scalar vec3_dot_product(const vec3 *a, const vec3 *b);
void   vec3_cross_product(vec3 *dst, const vec3 *a, const vec3 *b);
void   mat3_vec3_product(vec3 *dst, const mat3 *a, const vec3 *v);
void   mat3x4_vec3_product(vec3 *dst, const mat3x4 *a, const vec3 *v);
void   mat3x4_3x3_vec3_product(vec3 *dst, const mat3x4 *a, const vec3 *v);
void   mat3x4_column_0(vec3 *dst, const mat3x4 *a);
void   mat3x4_column_1(vec3 *dst, const mat3x4 *a);
void   mat3x4_column_2(vec3 *dst, const mat3x4 *a);
void   mat3x4_column_3(vec3 *dst, const mat3x4 *a);
void   mat3x4_3x3_line_0(vec3 *dst, const mat3x4 *a);
void   mat3x4_3x3_line_1(vec3 *dst, const mat3x4 *a);
void   mat3x4_3x3_line_2(vec3 *dst, const mat3x4 *a);

void   mat3_product(mat3 *dst, const mat3 *a, const mat3 *b);
void   mat3_add_product(mat3 *dst, const mat3 *a, const mat3 *b);
void   mat3_sub_product(mat3 *dst, const mat3 *a, const mat3 *b);
void   mat3x4_product(mat3x4 *dst, const mat3x4 *a, const mat3x4 *b);
void   mat4_product(mat4 *dst, const mat4 *a, const mat4 *b);

void   mat3_post_mult(mat3 *dst, const mat3 *b);
void   mat3x4_post_mult(mat3x4 *dst, const mat3x4 *b);
void   mat4_post_mult(mat4 *dst, const mat4 *b);
void   mat3_pre_mult(mat3 *dst, const mat3 *a);
void   mat3x4_pre_mult(mat3x4 *dst, const mat3x4 *a);
void   mat4_pre_mult(mat4 *dst, const mat4 *a);


// performs 3x3 MAM'
void   mat3_mat3x4_3x3_transform(mat3 *dst, const mat3x4 *m, const mat3 *a);

// inverses (Laplace)
scalar mat3_determinant(const mat3 *src);
void   mat3_inverse(mat3 *dst, const mat3 *src);
void   mat3_determined_inverse(mat3 *dst, const mat3 *src,
	const scalar detsrc);

scalar mat3x4_determinant(const mat3x4 *src);
void   mat3x4_inverse(mat3x4 *dst, const mat3x4 *src);
void   mat3x4_Rt_inverse(mat3x4 *dst, const mat3x4 *src);

scalar mat4_determinant(const mat4 *src);
void   mat4_inverse(mat4 *dst, const mat4 *src);


// debugging helpers
void scalar_print(const char *name, const scalar x);
void vec3_print(const char *name, const vec3 *v);
void vec4_print(const char *name, const vec4 *v);
void mat3_print(const char *name, const mat3 *a);
void mat3x4_print(const char *name, const mat3x4 *a);
void mat4_print(const char *name, const mat4 *a);

#define scalar_disp(a)	scalar_print(#a, a)
#define vec3_disp(a)	vec3_print(#a, a)
#define vec4_disp(a)	vec4_print(#a, a)
#define mat3_disp(a)	mat3_print(#a, a)
#define mat3x4_disp(a)	mat3x4_print(#a, a)
#define mat4_disp(a)	mat4_print(#a, a)

#endif

